import { IFrame, MainAreaWidget } from '@jupyterlab/apputils';
import { ServerConnection } from '@jupyterlab/services';
import { JSONObject } from '@lumino/coreutils';
import { ISignal } from '@lumino/signaling';
import { Widget } from '@lumino/widgets';
/**
 * Info for a a given dashboard URL.
 */
export declare type DashboardURLInfo = {
    /**
     * The user provided url in the search box.
     */
    url: string;
    /**
     * Whether there is a live dashboard at the URL.
     */
    isActive: boolean;
    /**
     * A new URL to use after redirects or proxies.
     */
    effectiveUrl?: string;
    /**
     * A mapping from individual dashboard plot names to their sub-path.
     */
    plots: {
        [name: string]: string;
    };
};
/**
 * A class for hosting a Dask dashboard in an iframe.
 */
export declare class DaskDashboard extends MainAreaWidget<IFrame> {
    /**
     * Construct a new dashboard widget.
     */
    constructor();
    /**
     * The current dashboard item for the widget.
     */
    get item(): IDashboardItem | null;
    set item(value: IDashboardItem | null);
    /**
     * The current dashboard URL for the widget.
     */
    get dashboardUrl(): string;
    set dashboardUrl(value: string);
    /**
     * Whether the dashboard is active. When inactive,
     * it will show a placeholder panel.
     */
    get active(): boolean;
    set active(value: boolean);
    /**
     * Handle an update request to the dashboard panel.
     */
    protected onUpdateRequest(): void;
    private _item;
    private _dashboardUrl;
    private _active;
    private _inactivePanel;
}
/**
 * A widget for hosting Dask dashboard launchers.
 */
export declare class DaskDashboardLauncher extends Widget {
    /**
     * Create a new Dask sidebar.
     */
    constructor(options: DaskDashboardLauncher.IOptions);
    private _updateLinks;
    /**
     * The list of dashboard items which can be launched.
     */
    get items(): IDashboardItem[];
    /**
     * Get the URL input widget.
     */
    get input(): URLInput;
    /**
     * Handle an update request.
     */
    protected onUpdateRequest(): void;
    /**
     * Rerender after showing.
     */
    protected onAfterShow(): void;
    private _dashboard;
    private _input;
    private _launchItem;
    private _items;
    private _serverSettings;
}
/**
 * A widget for hosting a url input element.
 */
export declare class URLInput extends Widget {
    /**
     * Construct a new input element.
     */
    constructor(serverSettings: ServerConnection.ISettings, linkFinder?: () => Promise<string>);
    /**
     * The underlying input value.
     */
    get input(): HTMLInputElement;
    /**
     * The base url for the dask webserver.
     *
     * #### Notes
     * Setting this value will result in a urlChanged
     * signal being emitted, but it will happen asynchronously,
     * as it first checks to see whether the url is pointing
     * at a valid dask dashboard server.
     */
    set url(newValue: string);
    /**
     * The URL information for the dashboard. This should be set via the url setter,
     * but read through this getter, as it brings in some extra information.
     */
    get urlInfo(): DashboardURLInfo;
    /**
     * A signal emitted when the url changes.
     */
    get urlInfoChanged(): ISignal<this, URLInput.IChangedArgs>;
    /**
     * Dispose of the resources held by the dashboard.
     */
    dispose(): void;
    /**
     * Handle the DOM events for the widget.
     *
     * @param event - The DOM event sent to the widget.
     *
     * #### Notes
     * This method implements the DOM `EventListener` interface and is
     * called in response to events on the main area widget's node. It should
     * not be called directly by user code.
     */
    handleEvent(event: KeyboardEvent): void;
    /**
     * Handle `after-attach` messages for the widget.
     */
    protected onAfterAttach(): void;
    /**
     * Handle `before-detach` messages for the widget.
     */
    protected onBeforeDetach(): void;
    /**
     * Periodically poll for valid url.
     */
    private _startUrlCheckTimer;
    private _urlChanged;
    private _urlInfo;
    private _input;
    private _poll;
    private _serverSettings;
}
/**
 * A namespace for URLInput statics.
 */
export declare namespace URLInput {
    /**
     * Changed args for the url.
     */
    interface IChangedArgs {
        /**
         * The old url info.
         */
        oldValue: DashboardURLInfo;
        /**
         * The new url info.
         */
        newValue: DashboardURLInfo;
    }
}
/**
 * A namespace for DaskDashboardLauncher statics.
 */
export declare namespace DaskDashboardLauncher {
    /**
     * Options for the constructor.
     */
    interface IOptions {
        /**
         * A function that attempts to find a link to
         * a dask bokeh server in the current application
         * context.
         */
        linkFinder?: () => Promise<string>;
        /**
         * A callback to launch a dashboard item.
         */
        launchItem: (item: IDashboardItem) => void;
        /**
         * A list of items for the launcher.
         */
        items?: IDashboardItem[];
    }
}
/**
 * Props for the dashboard listing component.
 */
export interface IDashboardListingProps {
    /**
     * A list of dashboard items to render.
     */
    items: IDashboardItem[];
    /**
     * A callback to launch a dashboard item.
     */
    launchItem: (item: IDashboardItem) => void;
    /**
     * Whether the items should be enabled.
     */
    isEnabled: boolean;
}
/**
 * An interface dashboard launcher item.
 */
export interface IDashboardItem extends JSONObject {
    /**
     * The route to add the the base url.
     */
    route: string;
    /**
     * The display label for the item.
     */
    label: string;
}
