import { IFrame, MainAreaWidget, ToolbarButton } from '@jupyterlab/apputils';
import { PageConfig, URLExt } from '@jupyterlab/coreutils';
import { ServerConnection } from '@jupyterlab/services';
import { searchIcon } from '@jupyterlab/ui-components';
import { JSONExt } from '@lumino/coreutils';
import { Poll } from '@lumino/polling';
import { Signal } from '@lumino/signaling';
import { Widget, PanelLayout } from '@lumino/widgets';
import * as React from 'react';
import * as ReactDOM from 'react-dom';
/**
 * A class for hosting a Dask dashboard in an iframe.
 */
export class DaskDashboard extends MainAreaWidget {
    /**
     * Construct a new dashboard widget.
     */
    constructor() {
        super({
            // Disable allow some iframe extensions to let server requests
            // and scripts to execute in the bokeh server context.
            // This is unsafe, but we presumably trust the code in the bokeh server.
            content: new IFrame({ sandbox: ['allow-scripts', 'allow-same-origin'] })
        });
        this._item = null;
        this._dashboardUrl = '';
        this._active = false;
        this._inactivePanel = Private.createInactivePanel();
        this.content.node.appendChild(this._inactivePanel);
        this.update();
    }
    /**
     * The current dashboard item for the widget.
     */
    get item() {
        return this._item;
    }
    set item(value) {
        if (JSONExt.deepEqual(value, this._item)) {
            return;
        }
        this._item = value;
        this.update();
    }
    /**
     * The current dashboard URL for the widget.
     */
    get dashboardUrl() {
        return this._dashboardUrl;
    }
    set dashboardUrl(value) {
        if (value === this._dashboardUrl) {
            return;
        }
        this._dashboardUrl = Private.normalizeDashboardUrl(value);
        this.update();
    }
    /**
     * Whether the dashboard is active. When inactive,
     * it will show a placeholder panel.
     */
    get active() {
        return this._active;
    }
    set active(value) {
        if (value === this._active) {
            return;
        }
        this._active = value;
        this.update();
    }
    /**
     * Handle an update request to the dashboard panel.
     */
    onUpdateRequest() {
        // If there is nothing to show, empty the iframe URL and
        // show the inactive panel.
        if (!this.item || !this.dashboardUrl || !this.active) {
            this.content.url = '';
            this._inactivePanel.style.display = '';
            return;
        }
        // Make sure the inactive panel is hidden
        this._inactivePanel.style.display = 'none';
        this.content.url = URLExt.join(this.dashboardUrl, this.item.route);
    }
}
/**
 * A widget for hosting Dask dashboard launchers.
 */
export class DaskDashboardLauncher extends Widget {
    /**
     * Create a new Dask sidebar.
     */
    constructor(options) {
        super();
        let layout = (this.layout = new PanelLayout());
        this._dashboard = new Widget();
        this._serverSettings = ServerConnection.makeSettings();
        this._input = new URLInput(this._serverSettings, options.linkFinder);
        layout.addWidget(this._input);
        layout.addWidget(this._dashboard);
        this.addClass('dask-DaskDashboardLauncher');
        this._items = options.items || [];
        this._launchItem = options.launchItem;
        this._input.urlInfoChanged.connect(this._updateLinks, this);
    }
    _updateLinks(_, change) {
        if (!change.newValue.isActive) {
            this.update();
            return;
        }
        const result = Private.getDashboardPlots(change.newValue);
        this._items = result;
        this.update();
    }
    /**
     * The list of dashboard items which can be launched.
     */
    get items() {
        return this._items;
    }
    /**
     * Get the URL input widget.
     */
    get input() {
        return this._input;
    }
    /**
     * Handle an update request.
     */
    onUpdateRequest() {
        // Don't bother if the sidebar is not visible
        if (!this.isVisible) {
            return;
        }
        ReactDOM.render(React.createElement(DashboardListing, { launchItem: this._launchItem, isEnabled: this.input.urlInfo.isActive, items: this._items }), this._dashboard.node);
    }
    /**
     * Rerender after showing.
     */
    onAfterShow() {
        this.update();
    }
}
/**
 * A widget for hosting a url input element.
 */
export class URLInput extends Widget {
    /**
     * Construct a new input element.
     */
    constructor(serverSettings, linkFinder) {
        super();
        this._urlChanged = new Signal(this);
        this._urlInfo = { isActive: false, url: '', plots: {} };
        this.addClass('dask-URLInput');
        const layout = (this.layout = new PanelLayout());
        const wrapper = new Widget();
        wrapper.addClass('dask-URLInput-wrapper');
        this._input = document.createElement('input');
        this._input.placeholder = 'DASK DASHBOARD URL';
        wrapper.node.appendChild(this._input);
        layout.addWidget(wrapper);
        this._serverSettings = serverSettings;
        if (linkFinder) {
            const findButton = new ToolbarButton({
                icon: searchIcon,
                onClick: async () => {
                    let link = await linkFinder();
                    if (link) {
                        this.url = link;
                    }
                },
                tooltip: 'Auto-detect dashboard URL'
            });
            layout.addWidget(findButton);
        }
        this._startUrlCheckTimer();
    }
    /**
     * The underlying input value.
     */
    get input() {
        return this._input;
    }
    /**
     * The base url for the dask webserver.
     *
     * #### Notes
     * Setting this value will result in a urlChanged
     * signal being emitted, but it will happen asynchronously,
     * as it first checks to see whether the url is pointing
     * at a valid dask dashboard server.
     */
    set url(newValue) {
        this._input.value = newValue;
        const oldValue = this._urlInfo;
        if (newValue === oldValue.url) {
            return;
        }
        void Private.testDaskDashboard(newValue, this._serverSettings).then(result => {
            this._urlInfo = result;
            this._urlChanged.emit({ oldValue, newValue: result });
            this._input.blur();
            this.update();
            if (!result) {
                console.warn(`${newValue} does not appear to host a valid Dask dashboard`);
            }
        });
    }
    /**
     * The URL information for the dashboard. This should be set via the url setter,
     * but read through this getter, as it brings in some extra information.
     */
    get urlInfo() {
        return this._urlInfo;
    }
    /**
     * A signal emitted when the url changes.
     */
    get urlInfoChanged() {
        return this._urlChanged;
    }
    /**
     * Dispose of the resources held by the dashboard.
     */
    dispose() {
        if (this.isDisposed) {
            return;
        }
        this._poll.dispose();
        super.dispose();
    }
    /**
     * Handle the DOM events for the widget.
     *
     * @param event - The DOM event sent to the widget.
     *
     * #### Notes
     * This method implements the DOM `EventListener` interface and is
     * called in response to events on the main area widget's node. It should
     * not be called directly by user code.
     */
    handleEvent(event) {
        switch (event.type) {
            case 'keydown':
                switch (event.keyCode) {
                    case 13: // Enter
                        event.stopPropagation();
                        event.preventDefault();
                        this.url = this._input.value;
                        break;
                    default:
                        break;
                }
                break;
            default:
                break;
        }
    }
    /**
     * Handle `after-attach` messages for the widget.
     */
    onAfterAttach() {
        this._input.addEventListener('keydown', this, true);
    }
    /**
     * Handle `before-detach` messages for the widget.
     */
    onBeforeDetach() {
        this._input.removeEventListener('keydown', this, true);
    }
    /**
     * Periodically poll for valid url.
     */
    _startUrlCheckTimer() {
        this._poll = new Poll({
            factory: async () => {
                const urlInfo = this._urlInfo;
                // Don't bother checking if there is no url.
                if (!urlInfo.url) {
                    return;
                }
                const result = await Private.testDaskDashboard(urlInfo.url, this._serverSettings);
                if (!result.isActive && urlInfo.isActive) {
                    console.warn(`The connection to dask dashboard ${urlInfo.url} has been lost`);
                }
                if (!JSONExt.deepEqual(result, urlInfo)) {
                    this._urlInfo = result;
                    this._urlChanged.emit({
                        oldValue: urlInfo,
                        newValue: result
                    });
                }
                // Throw an error if the connection died. If we don't raise/reject here,
                // then the poll won't back off.
                if (!result.isActive) {
                    throw Error(`No connection to ${urlInfo.url}`);
                }
            },
            frequency: { interval: 4 * 1000, backoff: true, max: 120 * 1000 },
            standby: 'when-hidden'
        });
        // When the URL changes, refresh the poll and backoff.
        this._urlChanged.connect(() => this._poll.refresh(), this);
    }
}
/**
 * A React component for a launcher button listing.
 */
function DashboardListing(props) {
    if (!props.isEnabled) {
        return (React.createElement("div", { className: "dask-DashboardListing-inactive" },
            React.createElement("span", { className: "dask-DashboardListing-inactive-title" }, "Dashboard not connected"),
            React.createElement("span", { className: "dask-DashboardListing-inactive-detail" }, "To connect, paste a dashboard URL in the box above, or create a new Dask cluster with the cluster manager below. If you are still unable to connect, check your network setup.")));
    }
    const items = [...props.items].sort((e1, e2) => e1.label <= e2.label ? -1 : 1);
    const listing = items.map(item => {
        return (React.createElement("li", { className: "dask-DashboardListing-item", key: item.route },
            React.createElement("button", { className: "jp-mod-styled jp-mod-accept", value: item.label, disabled: !props.isEnabled, onClick: () => props.launchItem(item) }, item.label)));
    });
    // Return the JSX component.
    return (React.createElement("div", null,
        React.createElement("ul", { className: "dask-DashboardListing-list" }, listing)));
}
/**
 * A namespace for private functionality.
 */
var Private;
(function (Private) {
    /**
     * Optionally remove a `status` route from a dashboard url.
     */
    function normalizeDashboardUrl(url, baseUrl = '') {
        if (isLocal(url)) {
            if (!baseUrl) {
                baseUrl = PageConfig.getBaseUrl();
            }
            // If the path-portion of the baseUrl has been included,
            // strip that off.
            const tmp = new URL(baseUrl);
            if (url.startsWith(tmp.pathname)) {
                url = url.slice(tmp.pathname.length);
            }
            // Fully qualify the local URL to remove any relative-path confusion.
            url = baseUrl + url;
        }
        // If 'status' has been included at the end, strip it.
        if (url.endsWith('status')) {
            url = url.slice(0, -'status'.length);
        }
        else if (url.endsWith('status/')) {
            url = url.slice(0, -'status/'.length);
        }
        return url;
    }
    Private.normalizeDashboardUrl = normalizeDashboardUrl;
    /**
     * Return the json result of /individual-plots.json
     */
    function getDashboardPlots(info) {
        const plots = [];
        for (let key in info.plots) {
            const label = key.replace('Individual ', '');
            const route = String(info.plots[key]);
            const plot = { route: route, label: label, key: label };
            plots.push(plot);
        }
        return plots;
    }
    Private.getDashboardPlots = getDashboardPlots;
    /**
     * Test whether a given URL hosts a dask dashboard.
     */
    async function testDaskDashboard(url, settings) {
        url = normalizeDashboardUrl(url, settings.baseUrl);
        // If this is a url that we are proxying under the notebook server,
        // check for the individual charts directly.
        if (url.indexOf(settings.baseUrl) === 0) {
            return ServerConnection.makeRequest(URLExt.join(url, 'individual-plots.json'), {}, settings)
                .then(async (response) => {
                if (response.status === 200) {
                    const plots = (await response.json());
                    return {
                        url,
                        isActive: true,
                        plots
                    };
                }
                else {
                    return {
                        url,
                        isActive: false,
                        plots: {}
                    };
                }
            })
                .catch(() => {
                return {
                    url,
                    isActive: false,
                    plots: {}
                };
            });
        }
        const response = await ServerConnection.makeRequest(URLExt.join(settings.baseUrl, 'dask', 'dashboard-check', encodeURIComponent(url)), {}, settings);
        const info = (await response.json());
        return info;
    }
    Private.testDaskDashboard = testDaskDashboard;
    function createInactivePanel() {
        const panel = document.createElement('div');
        panel.className = 'dask-DaskDashboard-inactive';
        return panel;
    }
    Private.createInactivePanel = createInactivePanel;
    /**
     * Test whether the url is a local url.
     *
     * #### Notes
     * This function returns `false` for any fully qualified url, including
     * `data:`, `file:`, and `//` protocol URLs.
     */
    function isLocal(url) {
        const { protocol } = URLExt.parse(url);
        return (url.toLowerCase().indexOf(protocol) !== 0 && url.indexOf('//') !== 0);
    }
    Private.isLocal = isLocal;
})(Private || (Private = {}));
